"""A list for converting phonetic representation in English to Ukrainian/Russian pronunciation
For clarity, an example word with its phonetic transcription including the commented sound is provided in the comments
The sound representation is based on the ARPAbet symbol set: http://www.speech.cs.cmu.edu/cgi-bin/cmudict"""
PHONEME2UK = {
    'AA': ['а', 'о'],           # odd:      [AA D]
    'AE': ['е', 'а'],           # at:       [AE T]
    'AH': ['а'],                # hut:	    [HH AH T]
    'AO': ['ау', 'оу'],         # ought:    [AO T]
    'AW': ['ау', 'ав'],         # cow:      [K AW]
    'AY': ['аі', 'ай'],         # hide	    [HH AY D]
    'B': ['б'],                 # be	    [B IY]
    'CH': ['ч'],                # cheese	[CH IY Z]
    'D': ['д'],                 # dee	    [D IY]
    'DH': ['з'],                # thee	    [DH IY]
    'EH': ['е', 'а'],           # Ed	    [EH D]
    'ER': ['йо', 'е'],          # hurt	    [HH ER T]
    'EY': ['ей', 'еі'],         # ate	    [EY T]
    'F': ['ф'],                 # fee	    [F IY]
    'G': ['г', 'ґ', 'к'],       # green	    [G R IY N]
    'HH': ['х'],                # he	    [HH IY]
    'IH': ['і'],                # it	    [IH T]
    'IY': ['і'],                # eat		[IY T]
    'JH': ['дж'],               # gee	    [JH IY]
    'K': ['к'],                 # key	    [K IY]
    'L': ['л'],                 # lee	    [L IY]
    'M': ['м'],                 # me	    [M IY]
    'N': ['н'],                 # knee	    [N IY]
    'NG': ['н', 'нг'],          # ping	    [P IH NG]
    'OW': ['оу', 'оа', 'о'],    # oat	    [OW T]
    'OY': ['ой', 'оі'],         # toy	    [T OY]
    'P': ['п'],                 # pee	    [P IY]
    'R': ['р'],                 # read	    [R IY D]
    'S': ['с'],                 # sea	    [S IY]
    'SH': ['ш', 'щ'],           # she	    [SH IY]
    'T': ['т'],                 # tea	    [T IY]
    'TH': ['з', 'с'],           # theta	    [TH EY T AH]
    'UH': ['у'],                # hood	    [HH UH D]
    'UW': ['у', 'ю'],           # two	    [T UW]
    'V': ['в'],                 # vee	    [V IY]
    'W': ['в', 'у'],            # we	    [W IY]
    'Y': ['й'],                 # yield	    [Y IY L D]
    'Z': ['з'],                 # zee		[Z IY]
    'ZH': ['ж'],                # seizure	[S IY ZH ER]
    ' ': [' ']
}


"A list of slavic letter combinations that can be represented with fewer characters"
COMBINEUK = {
    'йа': 'я',
    'йе': 'є',
    'йу': 'ю',
    'шч': 'щ',
}



from typing import List
import pandas as pd
from ...config import NUMERAL_CONVERTER_DATA_PATH
from .numeral_entry import NumeralEntry, Case, Gender, NumClass, Number
from .numeral_data import NumeralData
from ...numeral_preprocessor import preprocess_number_string
from ...utils import transform_to_morph_form


class NumeralDataLoader:
    """Class responsible for loading numeral data for different languages.

    This class handles loading numeral-related data from CSV files, including
    processing values based on their order, and storing the resulting data
    as instances of NumeralEntry.
    """

    # Cache for loaded languages
    __loaded_languages_cache: List[str] = []

    def load_language_data(self, lang: str):
        """Load numerical data for a specific language from a CSV file.

        Args:
            lang (str): The language code (e.g., 'en', 'uk').

        Raises:
            ValueError: If the file for the given language does not exist.

        This method reads the corresponding CSV file for the specified language,
        processes the data, and stores the numerical entries in a dictionary.
        """
        if not self.is_available_language(lang):
            raise ValueError(
                f'The specified language "{lang}" is not supported. '
                f'Please select one of the following: {", ".join(self.get_available_languages())}')

        filename = NUMERAL_CONVERTER_DATA_PATH / f'{lang}.csv'
        df = pd.read_csv(filename, sep=",", dtype={'order': int})
        df = df.apply(lambda col: col.map(lambda x: None if pd.isnull(x) else x))

        numeral_data = NumeralData()
        for i, row in df.iterrows():
            num_class = transform_to_morph_form(row.get('num_class'), NumClass)
            case = transform_to_morph_form(row.get('case'), Case)
            gender = transform_to_morph_form(row.get('gender'), Gender)
            number = transform_to_morph_form(row.get('number'), Number)
            value = int(row['value']) if row['order'] < 6 else 10 ** row['order']

            for string in row["string"].split(' '):
                if not string:
                    continue
                string = preprocess_number_string(string)

                idx = len(numeral_data) + 1
                numeral_data[idx] = NumeralEntry(
                    string=string,
                    order=row['order'],
                    num_class=num_class,
                    scale=row['scale'],
                    case=case,
                    gender=gender,
                    number=number,
                    value=value)

        return numeral_data

    def get_available_languages(self):
        """Return list of available languages by reading the directory.

        This method loads and caches the list of languages from the CSV filenames
        in the NUMERAL_CONVERTER_DATA_PATH folder.

        Returns:
            List[str]: List of language codes (e.g., ['en', 'uk']).

        This method ensures that the list of available languages is only loaded
        once and then cached for future calls.
        """
        if not self.__loaded_languages_cache:
            self.__loaded_languages_cache = [filename.stem for filename in NUMERAL_CONVERTER_DATA_PATH.glob('*.csv')]

        return self.__loaded_languages_cache

    def is_available_language(self, lang: str):
        """Check if a language is available in the data.

        Args:
            lang (str): The language code to check.

        Returns:
            bool: True if the language is available, False otherwise.

        This method checks if the language code is available in the list of
        loaded languages.
        """
        return lang in self.get_available_languages()



from dataclasses import dataclass
from typing import Dict

from .numeral_entry import NumeralEntry


@dataclass
class NumeralData(Dict[int, NumeralEntry]):
    """Class to represent numeral data that behaves like a dictionary."""

    is_available_case: bool = False
    is_available_gender: bool = False
    is_available_number: bool = False
    is_available_num_class: bool = False

    def __setitem__(self, idx: int, numeral_entry: NumeralEntry):

        super().__setitem__(idx, numeral_entry)
        if numeral_entry.number is not None:
            self.is_available_number = True
        if numeral_entry.gender is not None:
            self.is_available_gender = True
        if numeral_entry.num_class is not None:
            self.is_available_num_class = True
        if numeral_entry.case is not None:
            self.is_available_case = True



from typing import List
from dataclasses import dataclass

from ..flexi_trie import FlexiTrieNode
from .correction_detail import CorrectionDetail


@dataclass
class SearchState:
    position: int
    path: str
    node: FlexiTrieNode
    corrections: List[CorrectionDetail]



from typing import Tuple, Sequence

from .patterns import TOKEN_PATTERN


def tokenize(name: str) -> Tuple[Sequence[str], Sequence[str]]:
    """
    Retokenize tokens to string.

    ::Parameters::

    :param str name: input string

    :return Tuple[Sequence[str], Sequence[str]]: list of tokens and list of separators

    """
    tokenized = [(m.group(), m.span()) for m in TOKEN_PATTERN.finditer(name)]

    tokens = [text for text, span in tokenized]
    seps = [name[tokenized[i][1][1]: (tokenized[i+1][1][0] if i+1 < len(tokenized) else None)] for i in range(len(tokenized))]

    return tokens, seps


def detokenize(tokens: Sequence[str], seps: Sequence[str]) -> str:
    """
    Retokenize tokens to string.

    ::Parameters::

    :param Sequence[str] tokens: list of tokens
    :param Sequence[str] seps: list of separators

    :return str: output string

    """
    name = ''.join([(seps[i-1] if i else '') + tokens[i] for i in range(len(tokens))])
    if len(seps) == len(tokens):
        name += seps[len(tokens)-1]

    return name



from .index_builder import build_search_index, update_search_index
from .fuzzy_searcher import fuzzy_search, fuzzy_search_internal
from .search_index import SearchIndex, save_search_index, load_search_index


__all__ = (
    'build_search_index',
    'update_search_index',
    'fuzzy_search',
    'fuzzy_search_internal',
    'SearchIndex',
    'save_search_index',
    'load_search_index'
)



import re
from .patterns import ENGLISH_RULES, MULTIPLE_SPACES


def preprocess_numeral(numeral: str, lang: str) -> str:
    """Preprocess the numeral string based on language and general cleaning rules.

    This function processes the numeral string by:
    - Replacing hyphens with spaces for the English language.
    - Removing the word "and" between numerals for English.
    - Normalizing multiple spaces into a single space.
    - Converting the string to lowercase.

    Args:
        numeral (str): The numeral string to preprocess.
        lang (str): The language code (e.g., 'en' for English).

    Returns:
        str: The preprocessed numeral string.

    Example:
        >>> preprocess_numeral("three-and-four", "en")
        "three four"
    """
    # Handle specific language rules
    if lang == "en":
        numeral = ENGLISH_RULES.sub(" ", numeral)

    # General preprocessing (spaces and case normalization)
    numeral = MULTIPLE_SPACES.sub(" ", numeral).strip()  # Collapse multiple spaces into one
    numeral = numeral.lower()  # Convert to lowercase

    return numeral


def preprocess_number_string(number_string: str) -> str:
    """Preprocess the number string by stripping spaces and converting to lowercase.

    This function is used for basic number string preprocessing without language-specific logic.

    Args:
        number_string (str): The number string to preprocess.

    Returns:
        str: The preprocessed number string.

    Example:
        >>> preprocess_number_string("   FIVE   ")
        "five"
    """
    # Strips leading and trailing spaces and converts to lowercase
    number_string = number_string.strip()
    if number_string:
        number_string = number_string.lower()
    return number_string



from .search_engine import SearchEngine
from .correction import SymbolSubstitution, SymbolsDeletion, SymbolsTransposition, SymbolInsertion

__all__ = (
    'SearchEngine',
    'SymbolSubstitution',
    'SymbolsDeletion',
    'SymbolsTransposition',
    'SymbolInsertion'
)



from typing import Dict, List, Optional, Union
import dill

from .config import MAX_CORRECTION_RATE

from .utils import validate_symbol_weights
from .search_engine import SearchEngine
from .search_engine.fuzzy_search_result import FuzzySearchResult
from .flexi_trie import FlexiTrie


class FlexiDict:

    _trie: FlexiTrie
    _search_engine: SearchEngine
    _symbol_weights: Optional[Dict[str, float]] = None

    def __init__(
        self,
        search_engine: Optional[SearchEngine] = None,
        symbol_weights: Optional[Dict[str, float]] = None,
    ):

        self._trie = FlexiTrie()

        validate_symbol_weights(symbol_weights)
        self._symbol_weights = symbol_weights

        self._search_engine = search_engine or SearchEngine()

    def __setitem__(self, keyname: str, value_id: Union[int, str]):
        """
        Adds a value to the prefix tree under the specified key.

        Args:
            keyname (str): The key under which to store the value.
            value_id (int): The identifier of the value.

        """
        self._trie.add(keyname, value_id, self._symbol_weights)

    def __getitem__(self, query: str) -> Optional[Union[str, int]]:
        """
        Retrieves the first value associated with the specified key.

        Args:
            query (str): The key for which to retrieve the value.

        Returns:
            int: The first value associated with the specified key.

        Raises:
            TypeError: If `query` is not a string.
            KeyError: If the `query` key is not found in the dictionary.
        """
        exact_match_values = self._trie.find(query)
        if exact_match_values:
            return next(iter(exact_match_values))

        fuzzy_match_values = self.get(query)
        if fuzzy_match_values:
            return fuzzy_match_values[0]

        return None

    def get(self, query: str, max_correction_rate: Optional[float] = None) -> List[Union[str, int]]:

        if max_correction_rate is None:
            max_correction_rate = MAX_CORRECTION_RATE

        fuzzy_match_items = self._search_engine.search(
            trie=self.trie,
            query=query,
            max_correction_rate=max_correction_rate)
        return [item.value for item in fuzzy_match_items]

    def search(
            self,
            query: str,
            max_correction_rate: Optional[float] = None,
            max_correction_rate_for_leaves: Optional[float] = 1.
    ) -> List[Union[str, int]]:
        return [
            item.value
            for item in self.search_internal(
                query, max_correction_rate, max_correction_rate_for_leaves)]

    def search_internal(
            self,
            query: str,
            max_correction_rate: Optional[float] = None,
            max_correction_rate_for_leaves: Optional[float] = 1.) -> List[FuzzySearchResult]:

        fuzzy_match_items = self._search_engine.search(
            trie=self.trie,
            query=query,
            max_correction_rate=max_correction_rate,
            max_correction_rate_for_leaves=max_correction_rate_for_leaves
        )

        return fuzzy_match_items

    @property
    def trie(self):
        return self._trie

    def save_trie(self, filename: str):
        with open(filename, "wb") as f:
            dill.dump(file=f, obj=self._trie, recurse=True)

    def load_trie(self, filename: str) -> None:
        with open(filename, "rb") as f:
            self._trie = dill.load(file=f)



from typing import List, Tuple
from collections import deque

from .flexi_trie import FlexiTrieNode
from ..config import DEFAULT_TOPN_LEAVES


class FlexiTrieTraverser:
    """
    A utility class for traversing and interacting with a prefix tree (trie).

    This class provides methods to:
    - Apply a string to the tree and collect the nodes along the path.
    - Retrieve leaf nodes and their associated paths starting from a given node.

    Methods:
        apply_string(node, s): Applies a string to the prefix tree and returns the path of nodes traversed.
        get_node_leaves(node, topn): Collects up to `topn` leaf nodes and their paths from a given starting node.
    """

    @staticmethod
    def apply_string(node: FlexiTrieNode, s: str) -> List[FlexiTrieNode]:
        """
        Traverses the prefix tree using the given string, starting from the specified node.

        Args:
            node (FlexiTrieNode): The starting node of the traversal.
            s (str): The string to apply to the prefix tree.

        Returns:
            List[FlexiTrieNode]: A list of nodes traversed along the path determined by the string.
                                 If a character in the string does not exist in the tree, traversal stops, and the path is returned up to that point.

        Example:
            >>> root = FlexiTrieNode()
            >>> traverser = FlexiTrieTraverser()
            >>> nodes = traverser.apply_string(root, "test")
            >>> print(nodes)  # List of nodes representing the path for "test".
        """
        path_nodes: List[FlexiTrieNode] = []

        for char in s:
            node = node.get(char)
            if not node:
                return path_nodes

            path_nodes.append(node)

        return path_nodes

    @staticmethod
    def get_node_leaves(node: FlexiTrieNode, topn: int = DEFAULT_TOPN_LEAVES) -> List[Tuple[int, List[str]]]:
        """
        Collects leaf node values and their paths from the given starting node.

        Args:
            node (FlexiTrieNode): The starting node of the tree or subtree to begin the search.
            topn (int, optional): The maximum number of leaf nodes to retrieve. Defaults to `DEFAULT_TOPN_LEAVES`.

        Returns:
            List[Tuple[int, List[str]]]: A list of tuples where each tuple contains:
                                          - `int`: The value stored at the leaf node.
                                          - `List[str]`: The sequence of symbols (path) leading to the leaf node.

        Example:
            >>> root = FlexiTrieNode()
            >>> traverser = FlexiTrieTraverser()
            >>> leaves = traverser.get_node_leaves(root, topn=5)
            >>> print(leaves)  # [(value1, ['a', 'b']), (value2, ['c', 'd']), ...]

        Notes:
            - If the number of available leaves is less than `topn`, all leaves will be returned.
            - If the starting node has no leaves, the result will be an empty list.
        """
        queue = deque([(child, [key]) for key, child in node.children.items()])  # Queue holds pairs of (current_node, current_path)
        leaves = []

        while queue:
            current_node, current_path = queue.popleft()

            # If the current node is a leaf (it has values), add its values and paths
            if current_node.values:
                for value in current_node.values:
                    leaves.append((value, current_path))
                    if len(leaves) >= topn:
                        return leaves[:topn]

                if len(leaves) >= topn:
                    return leaves[:topn]

            # Add the children to the queue with the updated path
            queue.extend((child, current_path + [symbol]) for symbol, child in current_node.children.items())

        return leaves[:topn]



"""
Functions for transliteration from Latin to Cyrillic based on phonetic sounds.
"""
import logging
from typing import List
from itertools import product

from transliterate import translit

from .tokenizer import tokenize, detokenize
from .g2p_loader import get_g2p_model
from .preprocessor import _preprocess_cyrillic, _preprocess_phoneme
from .constants import PHONEME2UK


logger = logging.getLogger(__name__)


def latin_to_cyrillic_phonetic(s: str) -> List[str]:
    """
    Transliterates a Latin string into Cyrillic based on phonetic sounds.

    Args:
        s (str): Input string in Latin script.

    Returns:
        List[str]: A list of possible phonetic transliteration variants.

    Examples:
        >>> latin_to_cyrillic_phonetic('coca')
        ['кока']
        >>> latin_to_cyrillic_phonetic('cola')
        ['кола']
        >>> latin_to_cyrillic_phonetic('YARO Veggie cheese')
        ['єро веджі чіз', 'єро ваджі чіз', 'яро веджі чіз', 'яро ваджі чіз']
        >>> latin_to_cyrillic_phonetic('borjomi')
        ['бауроджія', 'боуроджія']
        >>> latin_to_cyrillic_phonetic('columbia')
        ['каламбія']
        >>> latin_to_cyrillic_phonetic('HOKKAIDO CLUB')
        ['хокаідо клаб', 'хокайдо клаб']
        >>> latin_to_cyrillic_phonetic('sonatural')
        ['санечел', 'саначел']
    """
    if not isinstance(s, str):
        raise TypeError("Input must be a string.")

    g2p = get_g2p_model()

    tokens, seps = tokenize(s)

    tokens_translit = []
    for token in tokens:
        phonemes = g2p(token) or [token, ]
        uk_letters = [PHONEME2UK.get(_preprocess_phoneme(phoneme), phoneme) for phoneme in phonemes]
        result = [_preprocess_cyrillic(''.join(x)) for x in product(*uk_letters)]
        result.append(translit(token, 'uk').lower())
        result.append(translit(token, 'ru').lower())
        tokens_translit.append(list(set(result)))

    result = [detokenize(tokens, seps) for tokens in product(*tokens_translit)]
    min_len = min(len(x) for x in result)

    print(s, result)
    return [x for x in result if len(x) <= min_len+2]



from typing import Tuple, Sequence

from .phonetic_translit.tokenizer.patterns import TOKEN_PATTERN


def tokenize(name: str) -> Tuple[Sequence[str], Sequence[str]]:
    """
    Retokenize tokens to string.

    ::Parameters::

    :param str name: input string

    :return Tuple[Sequence[str], Sequence[str]]: list of tokens and list of separators

    ::Example::

    >>> tokens, seps = tokenize('лосось/сьомга с/м 1-2кг')
    >>> tokens
    ... # ['лосось', '/', 'сьомга', 'с', '/', 'м', '1', '-', '2', 'кг']
    >>> seps
    ... # ['', '', ' ', '', '', ' ', '', '', '']

    """
    if not name.islower():
        name = name.lower()

    tokenized = [(m.group(), m.span()) for m in TOKEN_PATTERN.finditer(name)]

    tokens = [text for text, span in tokenized]
    seps = [name[tokenized[i][1][1]: tokenized[i+1][1][0]] for i in range(len(tokenized)-1)]

    return tokens, seps


def detokenize(tokens: Sequence[str], seps: Sequence[str]) -> str:
    """
    Retokenize tokens to string.

    ::Parameters::

    :param Sequence[str] tokens: list of tokens
    :param Sequence[str] seps: list of separators

    :return str: output string

    ::Example::

    >>> tokens = ['лосось', '/', 'сьомга', 'с', '/', 'м', '1', '-', '2кг']
    >>> seps = ['', '', ' ', '', '', ' ', '', '']
    >>> detokenize(tokens, seps)
    ... # 'лосось/сьомга с/м 1-2кг'

    """
    name = ''.join([(seps[i-1] if i else '') + tokens[i] for i in range(len(tokens))])

    return name



from .numeral_converter import numeral2int, int2numeral, int2numerals
from .numeral_converter_loader import get_available_languages, get_max_order, get_max_numeral_word_number
from .text import convert_numerical_in_text


__all__ = (
    'numeral2int', 'int2numeral', 'convert_numerical_in_text', 'int2numerals',
    'get_available_languages', 'get_max_order', 'get_max_numeral_word_number')




from typing import Type, Optional

from .numeral_data_collector.numeral_data_loader.numeral_entry import MorphForm


def transform_to_morph_form(input_string: str, morph_class: Type[MorphForm]) -> Optional[MorphForm]:
    """Converts a string into its corresponding morphological form based on the given enum class.

    Args:
        input_string (str): The string to be converted into the desired morphological form (e.g., "singular").
        morph_class (Type[MorphForm]): The enum class (e.g., `NumClass`, `Case`, `Gender`)
                                       that defines the target morph form.

    Returns:
        Optional[MorphForm]: The input string converted to the morphological form defined by the `morph_class` enum.

    """
    if not input_string:
        return None

    if isinstance(input_string, morph_class):
        return input_string




    try:
        if isinstance(input_string, str):
            input_string = input_string.lower()
        value = morph_class(input_string)
    except KeyError:
        raise ValueError(
            f'Invalid {morph_class.__name__} value "{input_string}"; '
            f'expected one of {[x.value for x in morph_class]}'
        )

    return value



from typing import Dict, Set, Optional
from dataclasses import dataclass

from flexi_dict import FlexiDict

from .numeral_data_loader.numeral_data import NumeralData


@dataclass
class NumeralDataContainer:
    numeral_data: NumeralData
    flexi_index: FlexiDict
    value_index: Dict[int, Set[int]]
    max_order: int
    max_numeral_word_number: Optional[int] = None

    def __post_init__(self):
        self.max_numeral_word_number = (self.max_order // 3 + 1) * 4



import re

from .constants import COMBINEUK
from .patterns import PHONEME_STRUCTURE_PATTERN


def _preprocess_cyrillic(s: str) -> str:
    """
    Preprocess Cyrillic text by replacing specific patterns.

    Args:
        s (str): Input string in Cyrillic script.

    Returns:
        str: Preprocessed string with applied replacements.
    """
    for pattern, replacement in COMBINEUK.items():
        s = re.sub(pattern, replacement, s)

    return s


def _preprocess_phoneme(phoneme: str) -> str:
    """
    Preprocess a phoneme by removing stress markers.

    Stress markers (0, 1, or 2) are ignored during transliteration. The function
    extracts the base phoneme using a regular expression.

    Args:
        phoneme (str): The input phoneme in ARPAbet format.

    Returns:
        str: The phoneme without the stress marker.
    """
    match = PHONEME_STRUCTURE_PATTERN.match(phoneme)
    if match:
        return match.group(1)

    return phoneme



from typing import Dict, Sequence, List, Tuple, Callable, Optional, Union
from itertools import product, chain
import re
import logging

from .tokenizer import tokenize, detokenize
from .config import MIN_START_TOKEN_LENGTH
from .phonetic_translit import latin_to_cyrillic_phonetic
from .search_index import SearchIndex


logger = logging.getLogger(__name__)


def build_search_index(
        data: List[Tuple[Union[int, str], str]],
        min_start_token_length: int = MIN_START_TOKEN_LENGTH,
        transliterate_latin: bool = False,
        callback: Optional[Callable] = None,
        should_stop: Optional[Callable] = None,
        callback_step: int = 10000
) -> SearchIndex:
    search_index = SearchIndex()
    update_search_index(
        search_index=search_index,
        data=data,
        min_start_token_length=min_start_token_length,
        transliterate_latin=transliterate_latin,
        callback=callback,
        should_stop=should_stop,
        callback_step=callback_step)
    return search_index


def update_search_index(
        search_index: SearchIndex,
        data: List[Tuple[int, str]],
        min_start_token_length: int = MIN_START_TOKEN_LENGTH,
        transliterate_latin: bool = False,
        callback: Optional[Callable] = None,
        should_stop: Optional[Callable] = None,
        callback_step: int = 10000
):

    for i, (idx, key) in enumerate(data):
        if i % callback_step == 0:
            if callback:
                callback(i / len(data))
            if should_stop and should_stop():
                break
        queries = __generate_queries(key, min_start_token_length, transliterate_latin)

        for rate, queries_data in queries.items():
            for query in queries_data:
                search_index[rate][query] = idx


def __generate_queries(
        key: str,
        min_start_token_length: int,
        transliterate_latin: bool
) -> Dict[int, Sequence[str]]:

    queries = dict()
    queries[0] = [key.lower(), ]

    tokens, seps = tokenize(key)

    if transliterate_latin:
        tokens_groups = [__get_latin_translit(token) for token in tokens]
        _queries = [detokenize(_tokens, seps) for _tokens in product(*tokens_groups)]
        queries[1] = [query for query in _queries if query != key]
    else:
        tokens_groups = [[token, ] for token in tokens]
    logger.debug(f'For key: "{key}" generated {len(tokens_groups)} tokens_groups: {tokens_groups}')

    queries[2] = []
    for i in range(1, len(tokens_groups)+1):
        alt_tokens_groups = chain(tokens_groups[i:], tokens_groups[:i])
        alt_seps = list(chain(seps[i:], [' ', ], seps[:i-1]))
        new_queries = [detokenize(_tokens, alt_seps) for _tokens in product(*alt_tokens_groups) if len(tokens[0]) >= min_start_token_length]
        queries[2].extend(new_queries)

    return queries


def __get_latin_translit(token: str) -> Sequence[str]:
    alias = [token, ]

    if re.match('[a-z ]+', token):
        alias.extend(latin_to_cyrillic_phonetic(token))

    return alias



from collections import OrderedDict
from typing import Optional, Set, Dict, Union


class FlexiTrie:
    """
    A prefix tree (trie) implementation that stores integer values and maintains ordering of children keys.

    Attributes:
        _node_count (int): The total number of nodes in the trie.
        _root (FlexiTrieNode): The root node of the trie.

    Methods:
        add(keys, value, symbol_weights): Adds a value to the trie along the path specified by `keys`.
        find(keys): Retrieves the set of values stored at the node specified by `keys`.
        __len__(): Returns the total number of nodes in the trie.
        __str__(): Returns a string representation of the trie structure.
    """

    def __init__(self):
        """Initializes the FlexiTrie with an empty root node and sets the node counter to 0."""
        self._node_count = 0
        self._root = FlexiTrieNode(self)

    def __len__(self) -> int:
        """Returns the total number of nodes in the trie."""
        return self._node_count

    def __str__(self) -> str:
        """Returns a string representation of the trie structure."""
        def dfs(node: 'FlexiTrieNode') -> str:
            children_s = ', '.join(
                [f'{k}⟶{list(child.values) or ""}{dfs(child)}' for k, child in node.children.items()]
            )
            return f'{{{children_s}}}' if children_s else ""

        return f'FlexiTrie(node_count={self._node_count}, trie={dfs(self._root)})'

    @property
    def node_count(self) -> int:
        """Returns the current number of nodes in the trie."""
        return self._node_count

    def inc_node_count(self):
        """Increments the node count by one."""
        self._node_count += 1

    def add(self, keys: str, value: int, symbol_weights: Optional[Dict[str, float]] = None):
        """
        Adds a value to the trie along the path specified by the sequence of keys.

        Args:
            keys (str): A string representing the path to the node where the value will be stored.
            value (int): The value to be stored at the final node.
            symbol_weights (Optional[Dict[str, float]]): Optional weights for ordering children keys.
        """
        node = self._root
        for key in keys:
            node = node.add_child(key, symbol_weights)
        node.add_value(value)

    def find(self, keys: str) -> Optional[Set[int]]:
        """
        Retrieves the set of values stored at the node specified by the path of keys.

        Args:
            keys (str): A string representing the path to the node.

        Returns:
            Optional[Set[int]]: The set of values stored at the final node, or None if the path does not exist.
        """
        node = self._root
        for key in keys:
            node = node.get(key)
            if node is None:
                return None
        return node.values

    @property
    def root(self):
        """Returns the root node of the trie."""
        return self._root


class FlexiTrieNode:
    """
    A node in the FlexiTrie structure.

    Attributes:
        _trie (FlexiTrie): Reference to the parent trie.
        _idx (int): A unique index identifying this node.
        children (OrderedDict[str, FlexiTrieNode]): Ordered dictionary of child nodes.
        values (Set[int]): Set of integer values stored at this node.

    Methods:
        add_child(key, symbol_weights): Adds a child node for the given key if it does not exist.
        get(key): Retrieves the child node corresponding to the given key.
        add_value(value): Adds a single integer value to the node.
        add_values(values): Adds multiple integer values to the node.
    """

    def __init__(self, trie: FlexiTrie):
        """Initializes a node with a reference to its parent trie."""
        self._trie = trie
        self._idx = self._trie.node_count
        self._trie.inc_node_count()
        self.children: OrderedDict[str, 'FlexiTrieNode'] = OrderedDict()
        self.values: Set[Union[int, str]] = set()

    def add_child(self, key: str, symbol_weights: Optional[Dict[str, float]] = None) -> 'FlexiTrieNode':
        """
        Adds a child node for the given key if it does not already exist.

        Args:
            key (str): The key for the child node.
            symbol_weights (Optional[Dict[str, float]]): Optional weights for ordering children keys.

        Returns:
            FlexiTrieNode: The child node corresponding to the given key.
        """
        if key not in self.children:
            self.children[key] = FlexiTrieNode(self._trie)
            if symbol_weights:
                self.children = OrderedDict(
                    sorted(self.children.items(), key=lambda x: -symbol_weights.get(x[0], 0))
                )
        return self.children[key]

    def get(self, key: str) -> Optional['FlexiTrieNode']:
        """
        Retrieves the child node for the given key.

        Args:
            key (str): The key of the child node.

        Returns:
            Optional[FlexiTrieNode]: The child node if it exists, or None if it does not exist.
        """
        return self.children.get(key)

    def __getitem__(self, key: str) -> 'FlexiTrieNode':
        """
        Allows indexing to retrieve child nodes.

        Args:
            key (str): The key of the child node.

        Returns:
            FlexiTrieNode: The child node corresponding to the key.

        Raises:
            KeyError: If the key does not exist.
        """
        node = self.get(key)
        if node is None:
            raise KeyError(f"Key '{key}' not found.")
        return node

    def add_value(self, value: int):
        """
        Adds a single integer value to this node.

        Args:
            value (int): The value to add.
        """
        if value not in self.values:
            self.values.add(value)

    def add_values(self, values: Set[int]):
        """
        Adds multiple integer values to this node.

        Args:
            values (Set[int]): A set of values to add.
        """
        self.values.update(values)

    def __str__(self) -> str:
        """Returns a string representation of the node."""
        return f'FlexiTrieNode(idx={self._idx}, values={self.values}, children={list(self.children.keys())})'

    @property
    def idx(self):
        """Returns the unique index of this node."""
        return self._idx

    @property
    def tree(self):
        """Returns the parent trie of this node."""
        return self._trie



import warnings
from typing import List, Optional, Dict, Set, Any
import math
from dataclasses import dataclass
import logging
from concurrent.futures import ThreadPoolExecutor

from .numeral_data_collector.flexi_index_builder.flexi_index_builder import FlexiDict
from .numeral_data_collector.numeral_data_loader.numeral_entry import NumeralEntry, Case, Gender, Number, NumClass
from .numeral_data_collector.numeral_data_loader.numeral_data import NumeralData
from .numeral_preprocessor import preprocess_numeral


logger = logging.getLogger(__name__)


@dataclass
class NumberItem:
    value: int
    order: int
    scale: bool


@dataclass
class NumeralWord:
    default: str
    alt: List[str]


def _numeral2number_items(
    numeral: str, lang: str, flexi_index: FlexiDict, numeral_data: NumeralData, multi_threaded: bool
) -> List[NumberItem]:
    """
    Converts a textual numeral into a list of NumberItem objects representing its components.

    Args:
        numeral (str): The textual representation of the numeral to convert.
        lang (str): The language of the numeral for preprocessing and interpretation.
        flexi_index (FlexiDict): A flexible dictionary-like structure for fuzzy searching numeral components.
        numeral_data (NumeralData): A data structure containing numeral definitions, values, and attributes.

    Returns:
        List[NumberItem]: A list of NumberItem objects representing the decomposed numeral.

    Raises:
        ValueError: If any part of the numeral cannot be converted to a number or contains invalid components.

    Example:
        >>> numeral = "one hundred twenty-three"
        >>> lang = "en"
        >>> flexi_index = FlexiDict()
        >>> numeral_data = load_numeral_data(lang)
        >>> result = _numeral2number_items(numeral, lang, flexi_index, numeral_data)
        >>> for item in result:
        >>>     print(item)
        NumberItem(value=1, order=2, scale=1)
        NumberItem(value=100, order=1, scale=0)
        NumberItem(value=23, order=0, scale=0)
    """
    if multi_threaded:
        return _numeral2number_items_multi_threaded(numeral, lang, flexi_index, numeral_data)

    numeral = preprocess_numeral(numeral, lang)
    number_items: List[NumberItem] = []

    for i, number_word in enumerate(reversed(numeral.split())):
        fuzzy_search_result = flexi_index.get(number_word)
        if not fuzzy_search_result:
            raise ValueError(f'Cannot convert "{number_word}" to integer')

        if i > 0:
            # Filter ordinal numerals from non-terminal positions
            fuzzy_search_result = [
                idx for idx in fuzzy_search_result if not _is_ordinal(numeral_data[idx])
            ]
            if not fuzzy_search_result:
                raise ValueError(f'Ordinal numeral word "{number_word}" inside numeral')

        numeral_entry = numeral_data[fuzzy_search_result[0]]
        number_items.append(NumberItem(
            value=numeral_entry.value,
            order=numeral_entry.order,
            scale=numeral_entry.scale,
        ))

    return number_items[::-1]


def __process_word(number_word: str, flexi_index: FlexiDict, numeral_data: NumeralData) -> dict:
    """Helper function to process a single word."""
    fuzzy_search_result = flexi_index.get(number_word)
    if not fuzzy_search_result:
        raise ValueError(f'Cannot convert "{number_word}" to integer')

    return {
        "word": number_word,
        "result": fuzzy_search_result,
        "entry": numeral_data[fuzzy_search_result[0]],
    }


def _numeral2number_items_multi_threaded(
    numeral: str, lang: str, flexi_index: FlexiDict, numeral_data: NumeralData
) -> List[NumberItem]:
    numeral = preprocess_numeral(numeral, lang)
    number_words = list(reversed(numeral.split()))

    results = []
    with ThreadPoolExecutor() as executor:
        # Map each word to its processing task
        future_to_word = {
            executor.submit(__process_word, word, flexi_index, numeral_data): word
            for word in number_words
        }
        for future in future_to_word:
            try:
                results.append(future.result())
            except ValueError as e:
                raise e

    # Post-process results to filter ordinals for non-terminal positions
    number_items = []
    for i, result in enumerate(results):
        number_word = result["word"]
        fuzzy_search_result = result["result"]

        if i > 0:
            # Filter ordinal numerals
            fuzzy_search_result = [
                idx for idx in fuzzy_search_result if not _is_ordinal(numeral_data[idx])
            ]
            if not fuzzy_search_result:
                raise ValueError(f'Ordinal numeral word "{number_word}" inside numeral')

        numeral_entry = numeral_data[fuzzy_search_result[0]]
        number_items.append(NumberItem(
            value=numeral_entry.value,
            order=numeral_entry.order,
            scale=numeral_entry.scale,
        ))

    return number_items[::-1]


def _number_items2int(number_items: List[NumberItem]) -> int:
    """
    Converts a list of NumberItem objects into an integer value.

    Args:
        number_items (List[NumberItem]): List of NumberItem objects, representing decomposed numeral parts.

    Returns:
        int: The integer value represented by the input NumberItem list.

    Raises:
        ValueError: If the structure of the number items is invalid or inconsistent.

    Example:
        >>> number_items = [
        >>>     NumberItem(value=1, order=2, scale=0),
        >>>     NumberItem(value=100, order=1, scale=0),
        >>>     NumberItem(value=23, order=0, scale=0),
        >>> ]
        >>> result = _number_items2int(number_items)
        >>> print(result)
        123
    """
    int_value = 0
    number_items = number_items[::-1]

    i_number = num_block_start = 0
    num_block_order = 0

    if number_items[0].scale:
        i_number = num_block_start = 1
        num_block_order = number_items[0].order

    while i_number < len(number_items):
        i_number, inner_order = __search_block(number_items, i_number, num_block_order)
        __check_correct_order(number_items, num_block_start, i_number, inner_order)
        __value = (
            _number_items2int(number_items[num_block_start:i_number][::-1])
            if inner_order
            else max(sum([x.value for x in number_items[num_block_start:i_number]]), 1)
        )

        int_value += (10**num_block_order) * __value
        if i_number >= len(number_items):
            return int(int_value)

        __check_number_is_correct_scale(number_items, i_number, int_value)
        num_block_order = number_items[i_number].order
        num_block_start = i_number + 1
        i_number += 1

    if num_block_start is not None:
        int_value += 10**num_block_order

    return int(int_value)


def _int2number_items(number: int, lang: str) -> List[NumberItem]:
    """
    Converts an integer into a list of NumberItem objects representing its components.

    Args:
        number (int): The integer to convert into number items.
        lang (str): The language of the numeral system (e.g., 'en' for English).

    Returns:
        List[NumberItem]: A list of NumberItem objects representing the decomposed number.

    Example:
        >>> _int2number_items(123, 'en')
        [NumberItem(value=1, order=2, scale=None),
         NumberItem(value=100, order=2, scale=None),
         NumberItem(value=23, order=0, scale=None)]

    Raises:
        None

    Notes:
        - Handles decomposition of numbers into units, tens, and hundreds.
        - Supports specific language-dependent behavior (e.g., English).
    """
    logger.debug(f'In _int2number_items:\n\tnumber: {number}\n\tlang: {lang}')

    # Special case: if the number is zero, return a single NumberItem with a value of 0.
    if number == 0:
        return [
            NumberItem(0, -1, None),
        ]

    number_items: List[NumberItem] = []  # Result list to store NumberItem objects.
    current_order, ones = 0, None  # `current_order` tracks the digit's position; `ones` stores the last "ones" digit.
    mem = None  # Temporary storage for higher order scales.

    while number:
        logger.debug(f'Step #{number}')
        digit = number % 10  # Extract the last digit of the number.

        if current_order % 3 == 2 and digit:  # Case for hundreds (order % 3 == 2).
            if mem:
                # Insert any stored scale from the previous block.
                logger.debug(f'Step #{number}: detected order % 3 == 2, inserted from mem: {mem}')
                number_items.append(mem)
                mem = None
            if lang == 'en':  # Language-specific handling for English.
                number_items.append(NumberItem(100, current_order % 3, None))
                number_items.append(NumberItem(digit, current_order % 3, None))
                logger.debug(f'Step #{number}: detected order == 2, inserted {digit} and {100}')
            else:  # General handling for non-English languages.
                number_items.append(NumberItem(100 * digit, current_order % 3, None))
                logger.debug(f'Step #{number}: detected order == 2, inserted {100 * digit}')

        elif current_order % 3 == 0:  # Case for units (order % 3 == 0).
            logger.debug(f'Step #{number}: detected order % 3 == 0')
            ones = digit  # Store the current "ones" digit.
            if current_order > 0:
                # Store the scale for this block.
                mem = NumberItem(10**current_order, current_order, True)

        else:  # Case for tens (order % 3 == 1).
            logger.debug(f'Step #{number}')
            if digit == 1 and ones > 0:  # Handle teens (e.g., 11–19).
                value = 10 * digit + ones
                if value:
                    if mem:
                        number_items.append(mem)
                        mem = None
                    number_items.append(NumberItem(value, current_order % 3, None))
                    logger.debug(f'Step #{number}: inserted {value}')
            else:  # Handle other tens or finalize stored "ones".
                if ones:
                    if mem:
                        number_items.append(mem)
                        mem = None
                    number_items.append(NumberItem(ones, 0, None))
                    logger.debug(f'Step #{number}: inserted {ones}')

                if digit:  # Handle non-zero tens.
                    if mem:
                        number_items.append(mem)
                        mem = None
                    number_items.append(NumberItem(10 * digit, current_order % 3, None))
                    logger.debug(f'Step #{number}: inserted {10 * digit}')

            ones = None  # Reset "ones" for the next iteration.

        current_order += 1  # Move to the next digit position.
        number //= 10  # Remove the last digit from the number.

    # Add the final "ones" digit if it exists.
    if ones:
        if mem:
            number_items.append(mem)
        number_items.append(NumberItem(ones, 0, None))

    # Handle the case where the last number item has a non-None scale.
    if number_items[-1].scale is not None:
        number_items.append(NumberItem(1, 0, None))

    logger.debug(f'Result: {number_items}')
    return number_items[::-1]  # Reverse the list to ensure correct order.


def __int2numeral_word(
        value: int,
        value_index: Dict[int, Set[int]],
        numeral_data: NumeralData,
        case: Optional[Case] = None,
        num_class: Optional[NumClass] = None,
        gender: Optional[Gender] = None,
        number: Optional[Number] = None) -> NumeralWord:

    case = case or Case.NOMINATIVE
    num_class = num_class or NumClass.CARDINAL
    gender = gender or Gender.MASCULINE
    number = number or (Number.SINGULAR if value in (0, 1) else Number.PLURAL)

    logger.debug(f'In __int2numeral_word(value={value}, case={case}, num_class={num_class}, gender={gender}, number={number})')

    sub_entry_indices = value_index[value]

    if len(sub_entry_indices) == 0:
        raise ValueError(f"no data for number {value}")

    numeral_words = list()

    if num_class and numeral_data.is_available_num_class:
        logger.debug(f'Check num_class, expect {num_class}, got in list: {[(numeral_data[x].string, numeral_data[x].num_class.value if numeral_data[x].num_class else "None") for x in sub_entry_indices]}')
        sub_entry_indices = [idx for idx in sub_entry_indices if numeral_data[idx].num_class is None or numeral_data[idx].num_class == num_class]
        logger.debug(f'Filtered list: {[numeral_data[x].string for x in sub_entry_indices]}')

    if case and numeral_data.is_available_case:
        logger.debug(f'Check case, expect {case}, got in list: {[(numeral_data[x].string, numeral_data[x].case.value if numeral_data[x].case else "None") for x in sub_entry_indices]}')
        sub_entry_indices = [idx for idx in sub_entry_indices if numeral_data[idx].case is None or numeral_data[idx].case == case]
        logger.debug(f'Filtered list: {[numeral_data[x].string for x in sub_entry_indices]}')

    if gender and numeral_data.is_available_gender:
        logger.debug(f'Check gender, expect {gender}, got in list: {[(numeral_data[x].string, numeral_data[x].gender.value if numeral_data[x].gender else "None") for x in sub_entry_indices]}')
        sub_entry_indices = [idx for idx in sub_entry_indices if numeral_data[idx].gender is None or numeral_data[idx].gender == gender]
        logger.debug(f'Filtered list: {[numeral_data[x].string for x in sub_entry_indices]}')

    if number and numeral_data.is_available_number:
        logger.debug(f'Check number, expect {number}, got in list: {[(numeral_data[x].string, numeral_data[x].number.value if numeral_data[x].number else "None") for x in sub_entry_indices]}')
        sub_entry_indices = [idx for idx in sub_entry_indices if numeral_data[idx].number is None or numeral_data[idx].number == number]
        logger.debug(f'Filtered list: {[numeral_data[x].string for x in sub_entry_indices]}')

    for idx in sub_entry_indices:
        numeral_entry = numeral_data[idx]
        if numeral_entry.string not in numeral_words:
            numeral_words.append(numeral_entry.string)
            logger.debug(f'+ added "{numeral_entry.string}" to number list')

    return NumeralWord(numeral_words[0], numeral_words[1:])


def _number_items2numeral(
        number_items: List[NumberItem],
        lang: str,
        numeral_data: NumeralData,
        value_index: Dict[int, Set[int]],
        case: Optional[Case] = None,
        num_class: Optional[NumClass] = None,
        gender: Optional[Gender] = None,
        number: Optional[Number] = None) -> Dict[str, Any]:

    logger.debug(
        f'In _number_items2numeral:\n'
        f'\t number_items: {[i.value for i in number_items]}\n'
        f'\t         case: {case}\n'
        f'\t    num_class: {num_class}\n'
        f'\t       gender: {gender}\n'
        f'\t       number: {number}')

    if num_class is not None and num_class == NumClass.COLLECTIVE and len(number_items) > 1:
        warnings.warn("Can't convert to collective numeral number; cardinal used")
        num_class = NumClass.CARDINAL

    numbers = list()
    for i, number_item in enumerate(number_items):
        logger.debug(f'\nNumber item #{i}/{len(number_items)} {number_item}')

        if i == len(number_items) - 1:

            d_case = __define_morph_case(case, number_items, i, num_class)
            d_number = __define_morph_number(number, number_items, i)
            logger.debug(
                f'\nNumber item #{i}/{len(number_items)}: detected last number; '
                f'next morph changes will be applied:\n\tcase: {case}->{d_case}\n\tnumber: {number}->{d_number}')
            numbers.append(
                __int2numeral_word(
                    number_item.value,
                    numeral_data=numeral_data,
                    value_index=value_index,
                    case=d_case,
                    number=d_number,
                    num_class=num_class,
                    gender=gender,
                )
            )
            continue

        if (
            (0 < number_item.value < 10)
            and (i + 1 < len(number_items))
            and number_items[i + 1].scale
        ):

            d_gender = __define_morph_gender(number_items, i)
            d_case = __define_morph_case(case, number_items, i, num_class)
            d_number = __define_morph_number(number, number_items, i)
            logger.debug(
                f'Number item #{i}/{len(number_items)}: detected inner number before scaled; '
                f'next morph changes will be applied:\n'
                f'\tgender: {gender}->{d_gender}\n\tcase: {case}->{d_case}\n\tnumber: {number}->{d_number}')
            numbers.append(
                __int2numeral_word(
                    number_item.value, case=d_case, gender=d_gender, numeral_data=numeral_data, number=d_number, value_index=value_index)
            )
            continue

        if number_item.scale:
            d_case = __define_morph_case(case, number_items, i, num_class)
            d_number = __define_morph_number(number, number_items, i)
            logger.debug(
                f'Number item #{i}/{len(number_items)}: Detected scaled; '
                f'next morph changes will be applied:\n\tnumber: {number}->{d_number}\n\tcase: {case}->{d_case}')
            numbers.append(
                __int2numeral_word(
                    number_item.value, numeral_data=numeral_data, value_index=value_index, case=d_case, number=d_number)
            )
            continue

        d_case = __define_morph_case(case, number_items, i, num_class)
        numbers.append(
            __int2numeral_word(number_item.value, numeral_data=numeral_data, value_index=value_index, case=d_case))

    return __process_numbers(numbers, number_items, lang=lang)


def __check_correct_order(
    number_items: List[NumberItem], start: int, end: int, inner_order: Optional[int]
):
    for k in range(start + 1, end):
        __order = (
            0
            if number_items[k].value % 10 ** number_items[k].order
            else number_items[k].order
        )

        if not inner_order and number_items[k - 1].order >= __order:
            raise ValueError(
                f"position {len(number_items) - k}: {number_items[k - 1].value}"
                f" with order {number_items[k - 1].order} stands after "
                f"{number_items[k].value} with less/equal order {__order}"
            )

        if inner_order and number_items[k - 1].order == __order:
            raise ValueError(
                f"position {len(number_items) - k}: {number_items[k - 1].value}"
                f" with order {number_items[k - 1].order} stands after "
                f"{number_items[k].value} with equal order {__order}"
            )


def __search_block(number_items, start, num_block_order):
    inner_order = None
    while start < len(number_items) and (
        not number_items[start].scale or number_items[start].order < num_block_order
    ):
        if number_items[start].scale and (
            inner_order is None or inner_order < number_items[start].order
        ):
            inner_order = number_items[start].order
        start += 1
    return start, inner_order


def __check_number_is_correct_scale(number_items, i_number, int_value):
    if not number_items[i_number].scale:
        raise ValueError(
            f"position {len(number_items) - 1 - i_number}: expects 10^(3n) or 100; "
            f"found {number_items[i_number].value}"
        )

    value_order = int(math.log10(int_value))
    if number_items[i_number].order <= value_order:
        raise ValueError(
            f"position {len(number_items) - 1 - i_number}: order of "
            f"{number_items[i_number].value}:{number_items[i_number].order} "
            f"is less/equal of summary order in next group: {value_order}"
        )


def __define_morph_number(global_number: str, number_items: List[NumberItem], i: int) -> Number:

    number = global_number

    prev_value = number_items[i - 1].value if i > 0 else 1

    if number_items[i].scale:
        number = Number.SINGULAR if prev_value == 1 else Number.PLURAL

    if i + 1 < len(number_items):
        if number_items[i + 1].scale and not number_items[i].scale:
            number = Number.PLURAL if number_items[i].value != 1 else Number.SINGULAR

    return number


def __define_morph_case(global_case, number_items, i, global_num_class) -> Case:
    logger.debug(f'In __define_morph_case:'
                 f'\n\t       case: {global_case}'
                 f'\n\t          i: {i}'
                 f'\n\tnumber_item: {number_items[i]}'
                 f'\n\t  num_class: {global_num_class}')

    case = global_case
    prev_value = number_items[i - 1].value if i > 0 else 1

    if i == len(number_items) - 1:
        logger.debug(
            f'In __define_morph_case: detect'
            f'\n\ti == len(number_items) - 1'
            f'\n\tprev_value: {prev_value}')
        if number_items[i].scale:
            if prev_value in (1, 2, 3, 4):
                logger.debug(
                    f'In __define_morph_case: detect'
                    f'\n\tprev_value in (1, 2, 3, 4)'
                    f'\n\tcase: {case} -> NOMINATIVE')
                case = Case.NOMINATIVE
            else:
                logger.debug(
                    f'In __define_morph_case: detect'
                    f'\n\tprev_value not in (1, 2, 3, 4)'
                    f'\n\tcase: {case} -> NOMINATIVE')
                case = Case.GENETIVE

        return case

    if global_num_class == NumClass.ORDINAL:
        logger.debug(
            f'In __define_morph_case: detect'
            f'\n\tglobal_num_class == NumClass.ORDINAL'
            f'\n\tcase: {case} -> NOMINATIVE')
        case = Case.NOMINATIVE

    if (
        (0 < number_items[i].value < 10)
        and (i + 1 < len(number_items))
        and number_items[i + 1].scale
    ):
        if case == Case.ACCUSATIVE:
            logger.debug(
                f'In __define_morph_case: detect'
                f'\n\tnext item is scaled'
                f'\n\tcase: {case}'
                f'\n\tcase: {case} -> NOMINATIVE')
            case = Case.NOMINATIVE
        return case

    if number_items[i].scale:
        logger.debug(
            f'In __define_morph_case: detect'
            f'\n\titem is scaled'
            f'\n\tcase: {case}')
        if case in (Case.NOMINATIVE, Case.ACCUSATIVE) or case is None:
            if prev_value in (1, 2, 3, 4):
                logger.debug(
                    f'In __define_morph_case: detect'
                    f'\n\tprev_value in (1, 2, 3, 4)'
                    f'\n\tcase: {case} -> NOMINATIVE ')
                case = Case.NOMINATIVE
            else:
                logger.debug(
                    f'In __define_morph_case: detect'
                    f'\n\tprev_value not in (1, 2, 3, 4)'
                    f'\n\tcase: {case} -> GENETIVE ')
                case = Case.GENETIVE

            return case

    if case == Case.ACCUSATIVE and i != len(number_items) - 2:
        case = Case.NOMINATIVE

    return case


def __define_morph_gender(number_items, i):
    return Gender.FEMININE if number_items[i + 1].value == 1000 else Gender.MASCULINE


def __process_numbers(numbers: List[NumeralWord], number_items, lang: str) -> Dict[str, Any]:
    if lang == "en":
        numbers__ = numbers.copy()
        numbers = list()
        i = 0
        while i < len(number_items):
            if (
                i + 1 < len(number_items)
                and number_items[i].order == 1
                and number_items[i + 1].order == 0
            ):
                numbers.append(
                    NumeralWord(
                        numbers__[i].default + "-" + numbers__[i + 1].default, []
                    )
                )
                i += 2
            else:
                numbers.append(numbers__[i])
                i += 1

    numeral = " ".join(
        [
            f"{number.default}" + (f" ({', '.join(number.alt)})" if number.alt else "")
            for number in numbers
        ]
    )

    numeral_forms = [
        " ".join(
            [
                (
                    [
                        numbers[i].default,
                    ]
                    + numbers[i].alt
                )[j]
                for i, j in enumerate(__combinations)
            ]
        )
        for __combinations in __get_combinations(
            *[range(1 + len(number.alt)) for number in numbers]
        )
    ]

    return {'numeral': numeral, 'numeral_forms': numeral_forms}


def _is_ordinal(numeral_entry: NumeralEntry):
    if numeral_entry.num_class is not None and numeral_entry.num_class == NumClass.ORDINAL:
        return True
    return False


def __get_combinations(*args):
    def __combinations(res, a):
        for r in res:
            for x in a:
                yield r + [
                    x,
                ]

    result = None
    for arr in args:
        if result is None:
            result = (
                [
                    _,
                ]
                for _ in arr
            )
        else:
            result = __combinations(result, arr)

    return result



from typing import Optional
from dataclasses import dataclass
from enum import Enum


class NumClass(str, Enum):
    """Enumeration for different types of numerals.

    Attributes:
        CARDINAL: Represents cardinal numbers (e.g., one, two, three).
        COLLECTIVE: Represents collective numerals (e.g., pair, dozen, двое).
        ORDINAL: Represents ordinal numbers (e.g., first, second, third).
    """
    CARDINAL = 'cardinal'      # en: one, two, three etc.
    COLLECTIVE = 'collective'  # en: pair, dozen; uk: двое, троє, сотня
    ORDINAL = 'ordinal'        # en: first, second, third, etc.


class Case(str, Enum):
    """Enumeration for grammatical cases in languages with declension.

    Attributes:
        ACCUSATIVE: Accusative case.
        DATIVE: Dative case.
        GENETIVE: Genitive case.
        INSTRUMENTAL: Instrumental case.
        NOMINATIVE: Nominative case.
        PREPOSITIONAL: Prepositional case.
    """
    ACCUSATIVE = 'accusative'
    DATIVE = 'dative'
    GENETIVE = 'genetive'
    INSTRUMENTAL = 'instrumental'
    NOMINATIVE = 'nominative'
    PREPOSITIONAL = 'prepositional'


class Gender(str, Enum):
    """Enumeration for grammatical genders.

    Attributes:
        FEMININE: Feminine gender.
        MASCULINE: Masculine gender.
        NEUTER: Neuter gender.
    """
    FEMININE = 'feminine'
    MASCULINE = 'masculine'
    NEUTER = 'neuter'


class Number(str, Enum):
    """Enumeration for singular and plural number forms.

    Attributes:
        PLURAL: Plural form.
        SINGULAR: Singular form.
    """
    PLURAL = 'plural'
    SINGULAR = 'singular'


@dataclass
class MorphForm:
    """Base class for all enumerations, ensuring they are logically related."""
    pass


@dataclass
class NumeralEntry:
    """Represents a numeral with its string representation, value, and grammatical attributes.

    Attributes:
        string (str): The string representation of the numeral (e.g., "one", "two").
        value (int): The numeric value of the numeral (e.g., 1, 2, 3).
        order (int): The order of the numeral in a sequence (e.g., 1 for first, 2 for second).
        num_class (NumClass): The class of the numeral (e.g., cardinal, ordinal).
        scale (bool): Whether the numeral is a scaled number (e.g., million, billion).
        case (Optional[Case]): The grammatical case of the numeral (optional).
        gender (Optional[Gender]): The grammatical gender of the numeral (optional).
        number (Optional[Number]): The number form (singular or plural) of the numeral (optional).
    """
    string: str
    value: int
    order: int
    scale: bool

    num_class: NumClass
    case: Optional[Case] = None
    gender: Optional[Gender] = None
    number: Optional[Number] = None



import certifi
import ssl
import logging

ssl._create_default_https_context = ssl._create_unverified_context
ssl._create_default_https_context = lambda: ssl.create_default_context(cafile=certifi.where())

from g2p_en import G2p


logger = logging.getLogger(__name__)


_G2P_MODEL: G2p = None


def get_g2p_model() -> G2p:
    global _G2P_MODEL
    if _G2P_MODEL is None:
        try:
            from nltk.corpus import cmudict
            import nltk
            cmudict.dict()
            nltk.data.find("averaged_perceptron_tagger_eng")
        except LookupError:
            import certifi
            import ssl
            import nltk

            ssl._create_default_https_context = ssl._create_unverified_context
            ssl._create_default_https_context = lambda: ssl.create_default_context(cafile=certifi.where())

            logger.info("Downloading required NLTK resource: cmudict...")
            nltk.download('cmudict')

            logger.info("Downloading required NLTK resource: averaged_perceptron_tagger_eng...")
            nltk.download('averaged_perceptron_tagger_eng')

        _G2P_MODEL = G2p()

    return _G2P_MODEL



import logging
from typing import Optional, Dict, Any
import math

from .numeral_data_collector.numeral_data_loader.numeral_entry import Case, Gender, Number, NumClass
from .numeral_converter_helpers import (
    _numeral2number_items,
    _int2number_items,
    _number_items2int,
    _number_items2numeral
)
from .numeral_converter_loader import _get_language_data, get_max_numeral_word_number, get_max_order
from .config import MAX_NUMERAL_LENGTH
from .utils import transform_to_morph_form


logger = logging.getLogger(__name__)


def numeral2int(numeral: str, lang: str, multi_threaded: bool = True) -> Optional[int]:
    """
    Converts the input numeral (in the form of a string) into an integer value for the given language.

    Args:
        numeral (str): The input numeral string (e.g., "forty two").
        lang (str): The language code (e.g., 'en', 'uk').

    Returns:
        Optional[int]: The corresponding integer value, or None if conversion fails.

    Example:
        >>> numeral2int("сорок два", lang="uk")
        42
    """
    language_data = _get_language_data(lang)

    length = len(numeral)
    if length > MAX_NUMERAL_LENGTH:
        raise ValueError(
            f'too long numeral; '
            f'expected numeral with less than {MAX_NUMERAL_LENGTH} symbols; got {length}.')

    words_number = len(numeral.split())
    max_order = get_max_order(lang)
    max_words_number = get_max_numeral_word_number(lang)
    if words_number > max_words_number:
        raise ValueError(
            f'For language "{lang}", only numbers of order less than or equal to "{max_order}" are supported; '
            f'expected numeral with less than {max_words_number} words.')

    logger.info(f"Converting numeral '{numeral}' to integer in language '{lang}'")

    number_items = _numeral2number_items(
        numeral=numeral, lang=lang,
        numeral_data=language_data.numeral_data,
        flexi_index=language_data.flexi_index,
        multi_threaded=multi_threaded
    )
    value = _number_items2int(number_items=number_items)
    return value


def int2numeral(
        value: int,
        lang: str,
        case: Optional[Case] = None,
        num_class: Optional[NumClass] = None,
        gender: Optional[Gender] = None,
        number: Optional[Number] = None) -> str:
    """
    Converts an integer to its corresponding numeral string in the given language and morphological form.

    Args:
        value (int): The input integer value (e.g., 42).
        lang (str): The language code (e.g., 'en', 'uk').
        case (Optional[Case]): The grammatical case to apply.
        num_class (Optional[NumClass]): The numerical class (e.g., ordinal, cardinal).
        gender (Optional[Gender]): The grammatical gender (e.g., masculine, feminine).
        number (Optional[Number]): The number form (singular or plural).

    Returns:
        str: The corresponding numeral string in the given morphological form.

    Example:
        >>> int2numeral(42, lang='uk', case="nominative", num_class="cardinal")
        'сорок два'
    """
    numeral = __int2numerals(value, lang, case, num_class, gender, number)
    return numeral['numeral']


def int2numerals(
        value: int,
        lang: str,
        case: Optional[Case] = None,
        num_class: Optional[NumClass] = None,
        gender: Optional[Gender] = None,
        number: Optional[Number] = None) -> str:
    """
    Converts an integer to its corresponding numeral string in the given language and morphological form.

    Args:
        value (int): The input integer value (e.g., 42).
        lang (str): The language code (e.g., 'en', 'uk').
        case (Optional[Case]): The grammatical case to apply.
        num_class (Optional[NumClass]): The numerical class (e.g., ordinal, cardinal).
        gender (Optional[Gender]): The grammatical gender (e.g., masculine, feminine).
        number (Optional[Number]): The number form (singular or plural).

    Returns:
        str: The corresponding numeral string in the given morphological form.

    Example:
        >>> int2numeral(42, lang='uk', case="nominative", num_class="cardinal")
        'сорок два'
    """
    numeral = __int2numerals(value, lang, case, num_class, gender, number)
    return numeral['numeral_forms']


def __int2numerals(
        value: int,
        lang: str,
        case: Optional[Case] = None,
        num_class: Optional[NumClass] = None,
        gender: Optional[Gender] = None,
        number: Optional[Number] = None) -> Dict[str, any]:

    language_data = _get_language_data(lang)

    if value != 0:
        order = int(round(math.log(value, 10)))
        max_order = get_max_order(lang)

        if order > max_order:
            raise ValueError(
                f'Numbers of order {order} are not supported. '
                f'For language "{lang}", only numbers of order less than or equal to {max_order} are supported.')

    number_items = _int2number_items(value, lang)

    numeral = _number_items2numeral(
        number_items,
        lang=lang,
        numeral_data=language_data.numeral_data,
        value_index=language_data.value_index,
        case=transform_to_morph_form(case, Case),
        num_class=transform_to_morph_form(num_class, NumClass),
        gender=transform_to_morph_form(gender, Gender),
        number=transform_to_morph_form(number, Number)
    )

    return numeral



import bisect
from typing import Optional

from flexi_dict.search_engine.fuzzy_search_result import FuzzySearchResult

from .search_index import SearchIndex
from .config import DEFAULT_QUERY_TRANSFORMATION_PRICE


def fuzzy_search(
        query: str,
        search_index: SearchIndex,
        topn: Optional[int] = None,
        query_transformation_price: float = DEFAULT_QUERY_TRANSFORMATION_PRICE
):
    return [
        item.value
        for item in fuzzy_search_internal(
            query=query,
            search_index=search_index,
            topn=topn,
            query_transformation_price=query_transformation_price)
    ]


def fuzzy_search_internal(
        query: str,
        search_index: SearchIndex,
        topn: Optional[int] = None,
        query_transformation_price: float = DEFAULT_QUERY_TRANSFORMATION_PRICE
):
    search_result = list()
    added = set()
    try:
        for rate in sorted(search_index):
            rate_search_result = search_index[rate].search_internal(
                query=query,
                max_correction_rate=.6,
                max_correction_rate_for_leaves=1.
            )
            for item in rate_search_result:

                if item.value in added:
                    continue

                added.add(item.value)
                if rate:
                    total_correction_price = __update_total_corrections_price(
                        item=item, rate=rate, query=query, query_transformation_price=query_transformation_price)
                    item.set_total_correction_price(total_correction_price)

                pos = bisect.bisect_left(
                    search_result, item.total_corrections_price, lo=0, hi=len(search_result),
                    key=lambda x: x.total_corrections_price)
                search_result.insert(pos, item)
                if topn and len(search_result) > topn:
                    raise StopIteration
    except StopIteration:
        pass

    return search_result


def __update_total_corrections_price(
        item: FuzzySearchResult, query: str, rate: int, query_transformation_price: float
):
    total_corrections_price = item.total_corrections_price
    if rate:
        total_corrections_price += query_transformation_price
        total_corrections_price += 1e-5 * __get_path_key_dist(item.path, item.key, query)

    return total_corrections_price


def __get_path_key_dist(path: str, key: str, query: str) -> float:
    """
    Посчитать цену разницы между длинами запроса и найденного пути для ранжирования результата поиска.

    Чем больше разница в длине - тем ниже опускаем в результате поиска.
    Ключ - это запрос с исправлениями, по которому нашли значение
    Путь - это путь в дереве до листка дереве, в котором лежит найденное значение

    Например, ключем может быть "ман" а путем - "мандарин" или "манка"

    :param str path: путь в дереве, по которому был найдено значение
    :param str key: ключ, по которому был найдено значение

    :return float: нормализированная цена разницы между длинами

    """
    # абсолютная разница между длиной пути и длиной ключа
    # если длина пути меньше длины запроса, то удваиваем путь;
    # например, если искали "ябл", а нашли "я" и "яблок", то "я" нужно опустиь ниже по рангу
    len_path, len_key, len_query = len(path), len(key), len(query)
    dist = len_path - len_key if len_path >= len_query else (len_query - len_path) ** 3

    # абсолютная разница между длиной последнего токена
    start = key.rfind(' ') + 1 if ' ' in key else 0
    len_path_last_token = (path[start:] + ' ').find(' ')
    len_key_last_token = len(key) - start
    len_token_dist = abs(len_path_last_token - len_key_last_token)

    # комбинируем расстояния
    if len_path >= len_query and not len_token_dist:
        dist = 0

    return dist



from typing import Optional, List
from dataclasses import dataclass, field

from .correction_detail import CorrectionDetail, get_total_corrections_price


@dataclass
class FuzzySearchResult:
    """
    Represents the result of a fuzzy search.

    Attributes:
        value (Optional[int]): The found value.
        key (str): The key corresponding to the result.
        path (str): The full path to the result in the prefix tree.
        corrections (Optional[List[CorrectionDetail]]): A list of corrections made during the search.
        total_corrections_price (float): The total "price" or score of all corrections.
    """
    value: Optional[int] = None
    key: str = ""
    path: str = ""
    corrections: Optional[List[CorrectionDetail]] = field(default_factory=list)
    total_corrections_price: float = 0.0

    def __post_init__(self):
        """
        Calculates the total correction price after the object is initialized.
        """
        self.total_corrections_price = get_total_corrections_price(self.corrections)

    def add_correction(self, correction: CorrectionDetail):
        self.corrections.append(correction)
        self.total_corrections_price = get_total_corrections_price(self.corrections)

    def set_total_correction_price(self, value: float):
        self.total_corrections_price = value



from setuptools import setup, find_packages

setup(
    name="flexi-nlp-tools",
    version="0.3.1",
    description="NLP toolkit based on the flexi-dict data structure, designed for efficient fuzzy search, with a focus on simplicity, performance, and flexibility.",
    author="Tetiana Lytvynenko",
    author_email="lytvynenkotv@gmail.com",
    license="MIT",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    keywords=["fuzzy search", "nlp tools", "flexi search", "numeral converter"],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=[
        "dill>=0.3.9,<0.4",
        "pandas>=2.2.3,<2.3",
        "g2p_en>=2.1.0, <2.2",
        "transliterate>=1.10.2, <1.11",
        "nltk>=3.9.1, <3.10",
        "certifi"
    ],
    python_requires=">=3.11",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
    package_data={
        "numeral_converter.resource": ["*.csv"],
    },
)



from typing import List, Optional
from dataclasses import dataclass


@dataclass
class CorrectionDetail:
    """
    Represents detailed information about a correction.

    Attributes:
        description (str): Correction description (e.g., for logging).
        position (Optional[int]): Position of the correction in the string (if applicable).
        price (Optional[float]): The "cost" or score of the correction.
    """
    correction_name: str
    parameters: List[str]
    position: Optional[int]
    price: Optional[float]


def get_total_corrections_price(corrections: Optional[List[CorrectionDetail]]) -> float:
    """
    Calculates the total "price" (score) of all corrections.

    Args:
        corrections (Optional[List[CorrectionDetail]]): A list of corrections applied to the query.

    Returns:
        float: The sum of correction scores. Returns 0.0 if no corrections are provided.
    """
    if not corrections:
        return 0.0

    return sum(correction.price for correction in corrections if correction.price is not None)



import logging
import bisect
from typing import List, Dict, Optional
from dataclasses import dataclass

from ..config import DEFAULT_TOPN_LEAVES, MAX_CORRECTION_RATE

from ..flexi_trie import FlexiTrie, FlexiTrieNode, FlexiTrieTraverser
from .fuzzy_search_result import FuzzySearchResult
from .correction_detail import CorrectionDetail
from .correction import Correction, SymbolInsertion, SymbolsDeletion, SymbolSubstitution, SymbolsTransposition
from ..config import MAX_QUEUE_SIZE

logger = logging.getLogger(__name__)


@dataclass
class SearchState:
    position: int
    path: str
    node: FlexiTrieNode
    corrections: List[CorrectionDetail]


class SearchEngine:
    """
    Search engine for performing fuzzy searches within an FlexiTrie.

    Attributes:
        _trie_traverser (FlexiTrieTraverser): Helper class to traverse the trie.
        _corrections (List[Correction]): List of correction strategies to apply during the search.
    """

    def __init__(self, corrections: Optional[List[Correction]] = None, symbol_insertion: Optional[Correction] = None):
        """
        Initializes the SearchEngine with a list of corrections.

        Args:
            corrections (List[Correction]): List of correction strategies.
        """
        self._trie_traverser = FlexiTrieTraverser()
        self._corrections = corrections or [SymbolInsertion(), SymbolsDeletion(), SymbolSubstitution(), SymbolsTransposition()]
        self._symbol_insertion = symbol_insertion or SymbolInsertion()

    def search(
            self, trie: FlexiTrie, query: str,
            topn_leaves: int = DEFAULT_TOPN_LEAVES,
            max_correction_rate: Optional[float] = None,
            max_correction_rate_for_leaves: Optional[float] = None
    ) -> List[FuzzySearchResult]:
        """
        Perform a fuzzy search in the trie based on the query.

        Args:
            trie (FlexiTrie): Trie structure to search.
            query (str): Query string.
            topn_leaves (int): Number of leaves to consider in results.
            max_correction_rate (float): Maximum allowed correction rate.

        Returns:
            List[FuzzySearchResult]: List of search results.
        """
        logger.info(f"Starting search for query: '{query}'")

        if max_correction_rate is None:
            max_correction_rate = MAX_CORRECTION_RATE
        if max_correction_rate_for_leaves is None:
            max_correction_rate_for_leaves = max_correction_rate

        search_result: List[FuzzySearchResult] = []
        visited_status: Dict[int, int] = {}
        added = set()

        path_nodes = self._trie_traverser.apply_string(node=trie.root, s=query)

        if len(path_nodes) == len(query):
            logger.info("Exact match found.")
            search_result = self._generate_search_result_from_node(
                node=path_nodes[-1],
                path=query,
                key=query,
                topn_leaves=topn_leaves,
                corrections=[],
                max_correction_rate_for_leaves=max_correction_rate_for_leaves
            )
            visited_status[path_nodes[-1].idx] = 1
            added = {x.value for x in search_result}

            if trie.find(query):
                return search_result
        queue = [SearchState(position + 1, query[:position + 1], node, []) for position, node in enumerate(path_nodes)]
        queue.append(SearchState(0, "", trie.root, []))

        iteration = 0

        while queue:
            iteration += 1
            logger.info(f"Iteration {iteration}: {len(queue)} states in the queue.")
            updated_queue = []

            for step in queue:
                logger.debug(f"Processing node ID {step.node.idx} at position {step.position}, path='{step.path}', corrections={[c.correction_name for c in step.corrections]}")

                if step.position == len(query):
                    node_result = self._generate_search_result_from_node(
                        node=step.node, key=query[:step.position], path=step.path,
                        topn_leaves=topn_leaves, corrections=step.corrections,
                        max_correction_rate_for_leaves=max_correction_rate_for_leaves)
                    for item in node_result:
                        if item.value not in added:
                            pos = bisect.bisect_left(
                                search_result, item.total_corrections_price,
                                lo=0, hi=len(search_result),
                                key=lambda x: x.total_corrections_price)
                            search_result.insert(pos, item)

                            added.add(item.value)
                            logger.info(f"Added result: path='{item.path}', value='{item.value}'")
                    continue

                if visited_status.get(step.node.idx) == 1:
                    logger.debug(f"Node ID {step.node.idx} already visited.")
                    continue

                visited_status[step.node.idx] = 1

                nodes_ids_path = self._trie_traverser.apply_string(step.node, query[step.position:])
                for i, node in enumerate(nodes_ids_path):
                    updated_queue.append(SearchState(
                        position=step.position + i + 1,
                        path=step.path + query[step.position: step.position + i + 1],
                        node=node,
                        corrections=step.corrections))
                    if len(updated_queue) > MAX_QUEUE_SIZE:
                        break
                if len(updated_queue) >= MAX_QUEUE_SIZE:
                    queue = updated_queue[:MAX_QUEUE_SIZE]
                    continue

                current_correction_rate = (len(step.corrections) + 1) / len(query)
                if current_correction_rate <= max_correction_rate:
                    for correction in self._corrections:
                        corrections_states = correction.apply(query=query, search_state=step)
                        for correction_state in corrections_states:
                            if correction_state.node.idx == step.node.idx:
                                visited_status[step.node.idx] = 0
                                logger.debug(f"Correction loopback detected for node ID {step.node.idx}. Marked for revisit.")
                        updated_queue.extend(corrections_states)
                        if len(updated_queue) > MAX_QUEUE_SIZE:
                            break
            if len(updated_queue) >= MAX_QUEUE_SIZE:
                queue = updated_queue[:MAX_QUEUE_SIZE]
                continue

            if not updated_queue:
                logger.info("No items left in queue. Search completed.")
                break

            queue = updated_queue

        logger.info(
            f"Search completed. Found {len(search_result)} results:\n" +
            "\n".join([
                f'{i+1}) "{x.path}" corr_rate: {len(x.corrections) / len(query):4.2f}, corr_price: {x.total_corrections_price:4.2f}, corrs: {x.corrections}'
                for i, x in enumerate(search_result)]))

        return search_result

    def _generate_search_result_from_node(
            self,
            node: FlexiTrieNode,
            key: str,
            path: str,
            topn_leaves: int,
            corrections: List[CorrectionDetail],
            max_correction_rate_for_leaves: float
    ) -> List[FuzzySearchResult]:
        """
        Generate search results from a trie node and its leaves.

        Args:
            node (FlexiTrieNode): Current node in the trie.
            key (str): Key string associated with the result.
            path (str): Path taken to reach this node.
            topn_leaves (int): Number of leaves to consider.
            corrections (List[CorrectionDetail]): List of corrections applied.

        Returns:
            List[FuzzySearchResult]: List of fuzzy search results.
        """
        logger.debug(f"Generating search result for node #{node.idx} with path: '{path}' and corrections: {corrections}")
        result = [
            FuzzySearchResult(value=value, key=key, path=path, corrections=corrections)
            for value in node.values
        ]
        leaves = self._trie_traverser.get_node_leaves(node, topn=topn_leaves)
        logger.debug(f'Generating search result for node #{node.idx}: found leaves: {leaves}')

        for value, leaf_path in leaves:
            leaf_path = ''.join(leaf_path)

            corrections = corrections + [
                    CorrectionDetail(
                        correction_name=self._symbol_insertion.name,
                        position=len(path) + c_pos,
                        parameters=[c, ],
                        price=self._symbol_insertion.price
                    )
                    for c_pos, c in enumerate(leaf_path)]
            correction_rate = min(1., len(corrections) / (len(path) + len(leaf_path)))
            logger.debug(
                f'Generating search result for node #{node.idx}: leaf "{"".join(leaf_path)}":\n'
                f'\t corrections number: {len(corrections)}\n'
                f'\t        path length: {len(path) + len(leaf_path)} ("{path + leaf_path}")\n'
                f'\t    correction rate: {correction_rate:4.2f}\n'
                f'\t        corrections: {[c.correction_name for c in corrections]}\n'
                f'\t    added to result: {correction_rate <= max_correction_rate_for_leaves}')

            if correction_rate > max_correction_rate_for_leaves:
                continue

            fuzzy_search_result = FuzzySearchResult(
                value=value,
                key=key,
                path=path + leaf_path,
                corrections=corrections
            )

            pos = bisect.bisect_left(
                                result, fuzzy_search_result.total_corrections_price,
                                lo=0, hi=len(result),
                                key=lambda x: x.total_corrections_price)
            result.insert(pos, fuzzy_search_result)

        logger.debug(f"Generated {len(result)} results for node #{node.idx}.")
        return result



from typing import Dict
from dataclasses import dataclass
from pathlib import Path

import dill

from flexi_dict import FlexiDict
from flexi_dict.search_engine.correction import SymbolInsertion
from flexi_dict.search_engine import SearchEngine


@dataclass
class SearchIndex(Dict[int, 'FlexiDict']):

    def __getitem__(self, key: int) -> 'FlexiDict':
        if key not in self:
            search_engine = SearchEngine(symbol_insertion=SymbolInsertion(price=.01))
            self[key] = FlexiDict(search_engine=search_engine)
        return super().__getitem__(key)


def save_search_index(search_index: SearchIndex, dirname: str):
    dirname = Path(dirname)
    dirname.mkdir(exist_ok=True)

    for k, v in search_index.items():
        with open(dirname / f'{k}.bin', "wb") as f:
            dill.dump(file=f, obj=v.trie, recurse=True)


def load_search_index(dirname: str) -> SearchIndex:
    dirname = Path(dirname)
    if not dirname.exists():
        raise FileNotFoundError(f'Directory "{dirname}" does not exist')
    filenames = sorted(dirname.glob('*.bin'), key=lambda x: x.stem)

    search_index = SearchIndex()

    search_engine = SearchEngine(symbol_insertion=SymbolInsertion(price=.01))
    for filename in filenames:
        flexi_dict = FlexiDict(search_engine=search_engine)
        flexi_dict.load_trie(filename)
        key = int(filename.stem)
        search_index[key] = flexi_dict

    return search_index
